; compiler.asm
;
; The compiler/interpreter.

; mode ( -- a )
; lookup ( name -- header | false )
; (compile) ( -- )
; process-word ( a -- true | false )
; process ( a -- true | false )
; header ( name -- )
; : ( -- )
; ; ( -- )

; mode ( -- a )
header code_mode,"mode",flag_none
code_mode:
        push    bx                      ; Push TOS.
        mov     bx,mode                 ; Address to mode becomes new TOS.
        next

; lookup ( name -- header | false )
header code_lookup,"lookup",flag_none
code_lookup:
        call    nest
        dw      code_nptr               ; nptr
.next:  dw      code_do_lit,nstack.end  ; nstack.end
        dw      code_over,code_eq       ; over =
        dw      code_do_cbranch,.not_found ; if equal ==> goto .not_found
        dw      code_duptor,code_fetch  ; dup>r @
        dw      code_over,code_swap     ; over swap
        dw      code_lookup_dict        ; lookup-dict
        dw      code_cdup               ; ?dup
        dw      code_do_cnbranch,.again ; if zero ==> goto .again
        dw      code_rto,code_drop      ; r> drop 
        dw      code_nip,code_exit      ; nip exit
.again: dw      code_rto,code_2plus     ; r> 2+
        dw      code_do_branch,.next    ; goto .next
.not_found:
        dw      code_2drop              ; 2drop
        dw      code_do_lit,0           ; 0
        dw      code_exit               ; exit

; lookup-dict ( name dict -- header | false )
header code_lookup_dict,"lookup-dict",flag_none
code_lookup_dict:
        call    nest
        dw      code_fetch              ; @
        dw      code_dup                ; dup
        dw      code_do_cnbranch,.not_found ; 0= if ==> goto .not_found
.next:  dw      code_2dup               ; 2dup          NOS = name searched.
        dw      code_do_lit,5,code_plus ; 5 +           TOS = header name.
        dw      code_seq                ; s=            Compare strings.
        dw      code_do_cbranch,.found  ; if equal ==> goto .found
        dw      code_fetch,code_dup     ; @ dup         Get next header.
        dw      code_do_cbranch,.next   ; if (TOS <> 0) ==> goto .next
.not_found:
        dw      code_2drop              ; 2drop
        dw      code_do_lit,0,code_exit ; 0 exit
.found:
        dw      code_nip,code_exit      ; nip exit

; (compile) ( -- )
header code_do_compile,"(compile)",flag_none
code_do_compile:
        lodsw                           ; ax = word to compile.
        mov     di,[dp]                 ; di = dp.
        add     word[dp],2              ; dp += 2.
        mov     [di],ax                 ; [di] = TOS.
        next

; process ( a -- true | false )
header code_process,"process",flag_none
code_process:
        call    nest
        dw      code_dup                ; dup
        dw      code_lookup             ; lookup
        dw      code_dup                ; dup
        dw      code_do_cnbranch,.false ; if TOS <> 0, goto .false
        dw      code_nip                ; nip
        dw      code_process_word       ; process-word
        dw      code_exit               ; exit
.false: dw      code_drop               ; drop
        dw      code_dup                ; dup
        dw      code_ston               ; s>n
        dw      code_do_cbranch,.number ; if TOS <> 0, goto .number
        dw      code_do_lit,.msg_error  ; .msg_error
        dw      code_print              ; print
        dw      code_print              ; print
        dw      code_cr                 ; cr
        dw      code_do_lit,0           ; 0
        dw      code_exit               ; exit
.number:
        dw      code_nip                ; nip
        dw      code_mode,code_fetch    ; mode @
        dw      code_do_lit,mode_interpret ; mode_interpret
        dw      code_eq                 ; =
        dw      code_do_cbranch,.number_interpret
.number_compile:
        dw      code_do_compile,code_do_lit
        dw      code_comma              ; ,
.number_interpret:
        dw      code_do_lit,-1          ; -1
        dw      code_exit               ; exit
.msg_error:
        db      0x12,0x0d,0x0a,"Word not found: "

; process-word ( a -- true | false )
header code_process_word,"process-word",flag_none
code_process_word:
        call    nest
        dw      code_mode,code_fetch    ; mode @
        dw      code_do_lit,mode_interpret ; mode_interpret
        dw      code_eq                 ; =
        dw      code_do_cbranch,.interpret ; if true goto .interpret
        dw      code_dup,code_do_lit,4  ; dup 4
        dw      code_plus               ; +       (get to flag byte)
        dw      code_cfetch             ; c@
        dw      code_do_lit,flag_immediate ; flag_immediate
        dw      code_and                ; and
        dw      code_do_cbranch,.interpret ; if true goto .interpret
.compile:
        dw      code_2plus              ; 2+
        dw      code_fetch              ; @       (get code address)
        dw      code_comma              ; ,       (compile the word)
        dw      code_do_lit,-1          ; -1
        dw      code_exit               ; exit
.interpret:
        dw      code_2plus              ; 2+
        dw      code_fetch              ; @       (get code address)
        dw      code_execute            ; execute
        dw      code_do_lit,-1          ; -1
        dw      code_exit               ; exit

; header ( name -- )
header code_header,"header",flag_none
code_header:
        call    nest
        dw      code_dp,code_last       ; dp last
        dw      code_comma              ; ,        (store ptr to last)
        dw      code_laststore          ; last!    (make last point to this)
        dw      code_dup,code_cfetch    ; dup c@   (get length)
        dw      code_do_lit,6,code_plus ; 6 +      (dp+2, but ptr to header
        dw      code_dp,code_plus       ; dp +      is before code)
        dw      code_comma              ; ,        (compile ptr to code)
        dw      code_do_lit,0           ; 0
        dw      code_ccomma             ; c,       (compile flags)
        dw      code_scomma             ; s,
        dw      code_last,code_comma    ; last ,
        dw      code_exit               ; exit

; : ( -- )
header code_colon,":",flag_none
code_colon:
        call    nest
        dw      code_read,code_word     ; read word
        dw      code_header             ; header
        dw      code_do_lit,0xe8        ; 0xe8      (call instruction)
        dw      code_ccomma             ; c,
        dw      code_do_lit,nest        ; nest
        dw      code_dp                 ; last 2+ @  <--- ?
        dw      code_minus,code_2minus  ; here - 2-
        dw      code_comma              ; ,
        dw      code_do_lit,mode_compile ; mode_compile
        dw      code_mode,code_store    ; mode !
        dw      code_exit               ; exit

; ; ( -- )
header code_semicol,";",flag_immediate
code_semicol:
        call    nest
        dw      code_do_lit,mode_interpret
        dw      code_mode,code_store    ; mode_interpret mode !
        dw      code_do_compile,code_exit ; compile exit
        dw      code_exit               ; exit
