; core.asm
;
; This is where the core language is stored. Don't place more than
; necessary here.

; nest ( -- )
; exit ( -- )
; execute ( a -- )
; bye ( -- )
; (;code)
; (lit) ( -- a )
; (ref) ( -- )
; (constant) ( -- a )
; (variable) ( -- a )
; r@ ( -- a )
; >r ( a -- )
; dup>r ( a -- a )
; r> ( a -- )
; drop ( a -- )
; 2drop ( a a -- )
; dup ( a -- a a )
; ?dup ( a -- a ) or ( a -- a a ) if a <> 0
; 2dup ( a -- a a )
; swap ( a b -- b a )
; over ( a b -- a b a )
; pick ( a -- b )
; tuck ( a b -- b a b )
; nip ( a b -- a )
; rot ( a b c -- b c a )
; -rot ( a b c -- c a b )
; and ( a b -- a and b )
; or ( a b -- a or b )
; xor ( a b -- a xor b )
; + ( a b -- a+b )
; - ( a b -- a-b )
; * ( a b -- a*b )
; / ( a b -- a/b )
; mod ( a b -- a modulo b )
; */ ( a b c -- a*b / c )
; << ( a b -- a<<b )
; >> ( a b -- a>>b )
; rol ( a b -- c )
; ror ( a b -- c )
; 1+ ( a -- a+1 )
; 2+ ( a -- a+2 )
; 3+ ( a -- a+3 )
; 2* ( a -- a*2 )
; 2/ ( a -- a/2 )
; 1- ( a -- a-1 )
; 2- ( a -- a-2 )
; neg ( a -- -a )
; not ( a -- a )
; 0= ( a -- true | false )
; = ( a b -- true | false )
; <> ( a b -- true | false )
; < ( a b -- true | false )
; > ( a b -- true | false )
; @ ( a -- [a] )
; c@ ( a -- [a] )
; ! ( a b -- )
; +! ( a b -- )
; c! ( a b -- )
; (branch) ( -- )
; (?branch) ( a -- )
; (?nbranch) ( a -- )
; dp ( -- a )
; dp! ( a -- )
; last ( -- a )
; last! ( a -- )
; , ( a -- )
; c, ( a -- )
; (
; (next)
; param ( -- a )
; (does)
; sp0 ( -- n )
; sp ( -- n )
; sp! ( n -- )
; rp0 ( -- n )
; rp ( -- n )
; rp! ( n -- )
; cs ( -- cs )


header nest,"nest",flag_none
nest:
        xchg    bp,sp                   ; Swap return and stack pointers.
        push    si                      ; Push IP to return stack.
        xchg    bp,sp                   ; Swap return and stack pointers.
        pop     si                      ; Get IP from hardware stack.
        next                            ; Start executing thread.

; exit ( -- )
header code_exit,"exit",flag_none
code_exit:
        xchg    bp,sp
        pop     si
        xchg    bp,sp                   ; Pop IP from return stack.
        next

; execute ( a -- )
header code_execute,"execute",flag_none
code_execute:
        xchg    ax,bx                   ; ax = TOS.
        pop     bx                      ; Pop new TOS.
        jmp     ax                      ; Jump to code at [ax].

; bye ( -- )
header code_bye,"bye",flag_none
code_bye:
        int     0x20

; (;code)
header code_do_code,"(;code)",flag_none
code_do_code:
        jmp     si

; (lit) ( -- a )
header code_do_lit,"(lit)",flag_none
code_do_lit:
        push    bx                      ; Push old TOS.
        lodsw                           ; ax = word pointed to by IP.
        xchg    ax,bx                   ; Push the word.
        next

; (ref) ( -- )
header code_do_ref,"(ref)",flag_none
code_do_ref:
        pop     di
        jmp     [di]

; (constant) ( -- a )
header code_do_constant,"(constant)",flag_none
code_do_constant:
        pop     di                      ; di now points to the constant.
        push    bx                      ; Now push bx so it becomes NOS.
        mov     bx,[di]                 ; bx is the constant.
        next

; (variable) ( -- a )
header code_do_variable,"(variable)",flag_none
code_do_variable:
        xchg    ax,bx
        pop     bx
        push    ax
        next

; r@ ( -- a )
header code_rfetch,"r@",flag_none
code_rfetch:
        push    bx
        mov     bx,[bp]
        next

; >r ( a -- )
header code_tor,">r",flag_none
code_tor:
        xchg    bp,sp
        push    bx
        xchg    bp,sp
        pop     bx
        next

; dup>r ( a -- a )
header code_duptor,"dup>r",flag_none
code_duptor:
        xchg    bp,sp
        push    bx
        xchg    bp,sp
        next

; r> ( a -- )
header code_rto,"r>",flag_none
code_rto:
        push    bx
        xchg    bp,sp
        pop     bx
        xchg    bp,sp
        next

; drop ( a -- )
header code_drop,"drop",flag_none
code_drop:
        pop     bx                      ; Load new TOS.
        next

; 2drop ( a a -- )
header code_2drop,"2drop",flag_none
code_2drop:
        pop     ax                      ; Drop NOS.
        pop     bx                      ; Load new TOS.
        next

; dup ( a -- a a )
header code_dup,"dup",flag_none
code_dup:
        push    bx                      ; Push TOS to hardware stack.
        next

; ?dup ( a -- a ) or ( a -- a a ) if a <> 0
header code_cdup,"?dup",flag_none
code_cdup:
        or      bx,bx                   ; TOS = 0?
        jz      .zero                   ; Yes, ship "dup".
        push    bx                      ; Push TOS to hardware stack.
.zero:  next

; 2dup ( a -- a a )
header code_2dup,"2dup",flag_none
code_2dup:
        pop     ax                      ; ax = NOS.
        push    ax                      ; NOS in the bottom.
        push    bx                      ; Then TOS.
        push    ax                      ; Then NOS again.
        next

; swap ( a b -- b a )
header code_swap,"swap",flag_none
code_swap:
        pop     ax                      ; ax = NOS.
        xchg    ax,bx                   ; Swap ax and bx (TOS).
        push    ax                      ; Make old TOS new NOS.
        next

; over ( a b -- a b a )
header code_over,"over",flag_none
code_over:
        pop     ax                      ; ax = NOS.
        push    ax                      ; Old NOS on the bottom.
        push    bx                      ; Old TOS next.
        xchg    ax,bx                   ; Old NOS on top.
        next

; pick ( a -- b )
header code_pick,"pick",flag_none
code_pick:
        mov     di,sp                   ; di = pointer to NOS
        shl     bx,1
        mov     bx,[bx+di]              ; TOS = stack[TOS*2]
        next

; tuck ( a b -- b a b )
header code_tuck,"tuck",flag_none
code_tuck:
        pop     ax                      ; ax = NOS.
        push    bx                      ; TOS on bottom.
        push    ax                      ; NOS = ax.
        next

; nip ( a b -- a )
header code_nip,"nip",flag_none
code_nip:
        pop     ax
        next

; rot ( a b c -- b c a )
header code_rot,"rot",flag_none
code_rot:
        xchg    sp,bp
        xchg    bx,[bp]
        xchg    bx,[bp+2]
        xchg    sp,bp
        next

; -rot ( a b c -- c a b )
header code_mrot,"-rot",flag_none
code_mrot:
        xchg    sp,bp
        xchg    bx,[bp+2]
        xchg    bx,[bp]
        xchg    sp,bp
        next

; and ( a b -- a and b )
header code_and,"and",flag_none
code_and:
        pop     ax                      ; ax = NOS.
        and     bx,ax                   ; TOS = TOS and NOS.
        next

; or ( a b -- a or b )
header code_or,"or",flag_none
code_or:
        pop     ax                      ; ax = NOS.
        or      bx,ax                   ; TOS = TOS or NOS.
        next

; xor ( a b -- a xor b )
header code_xor,"xor",flag_none
code_xor:
        pop     ax                      ; ax = NOS.
        xor     bx,ax                   ; TOS = TOS or NOS.
        next

; + ( a b -- a+b )
header code_plus,"+",flag_none
code_plus:
        pop     ax                      ; ax = NOS.
        add     bx,ax                   ; TOS = TOS - NOS.
        next

; - ( a b -- a-b )
header code_minus,"-",flag_none
code_minus:
        pop     ax                      ; ax = NOS.
        sub     bx,ax                   ; 
        neg     bx                      ; TOS = NOS - TOS.
        next

; * ( a b -- a*b )
header code_mul,"*",flag_none
code_mul:
        pop     ax                      ; ax = NOS.
        mul     bx                      ; dx:ax = NOS * TOS.
        xchg    ax,bx                   ; bx = ax.
        next

; / ( a b -- a/b )
header code_div,"/",flag_none
code_div:
        pop     ax                      ; ax = NOS.
        xor     dx,dx                   ; dx:ax = ax.
        div     bx                      ; ax = dx:ax / bx.
        xchg    ax,bx                   ; TOS = ax.
        next

; mod ( a b -- a modulo b )
header code_mod,"mod",flag_none
code_mod:
        pop     ax                      ; ax = NOS.
        xor     dx,dx                   ; dx:ax = ax.
        div     bx                      ; dx = dx:ax modulo bx.
        mov     bx,dx                   ; TOS = dx.
        next

; */ ( a b c -- a*b / c )
header code_muldiv,"*/",flag_none
code_muldiv:
        pop     ax                      ; ax = NOS.
        pop     cx                      ; cx = 3rd pos.
        mul     cx                      ; dx:ax = ax*cx.
        div     bx                      ; ax = dx:ax / bx.
        xchg    ax,bx                   ; TOS = ax.
        next

; << ( a b -- a<<b )
header code_shl,"<<",flag_none
code_shl:
        mov     cl,bl                   ; cl = TOS.
        pop     bx                      ; TOS = NOS.
        shl     bx,cl                   ; TOS = TOS << cl.
        next

; >> ( a b -- a>>b )
header code_shr,">>",flag_none
code_shr:
        mov     cl,bl                   ; cl = TOS.
        pop     bx                      ; TOS = NOS.
        shr     bx,cl                   ; TOS = TOS >> cl.
        next

; rol ( a b -- c )
header code_rol,"rol",flag_none
code_rol:
        mov     cl,bl
        pop     bx
        rol     bx,cl
        next

; ror ( a b -- c )
header code_ror,"ror",flag_none
code_ror:
        mov     cl,bl
        pop     bx
        ror     bx,cl
        next

; 1+ ( a -- a+1 )
header code_1plus,"1+",flag_none
code_1plus:
        inc     bx                      ; TOS = TOS + 1.
        next

; 2+ ( a -- a+2 )
header code_2plus,"2+",flag_none
code_2plus:
        add     bx,2                    ; TOS = TOS + 2.
        next

; 3+ ( a -- a+3 )
header code_3plus,"3+",flag_none
code_3plus:
        add     bx,3                    ; TOS = TOS + 3.
        next

; 2* ( a -- a*2 )
header code_2mul,"2*",flag_none
code_2mul:
        shl     bx,1                    ; TOS = TOS * 2.
        next
  
; 2/ ( a -- a/2 )
header code_2div,"2/",flag_none
code_2div:
        shr     bx,1                    ; TOS = TOS / 2.
        next

; 1- ( a -- a-1 )
header code_1minus,"1-",flag_none
code_1minus:
        dec     bx                      ; TOS = TOS - 1.
        next

; 2- ( a -- a-2 )
header code_2minus,"2-",flag_none
code_2minus:
        sub     bx,2                    ; TOS = TOS - 2.
        next

; neg ( a -- -a )
header code_neg,"neg",flag_none
code_neg:
        neg     bx                      ; TOS = - TOS.
        next

; not ( a -- a )
header code_not,"not",flag_none
code_not:
        not     bx                      ; TOS = not TOS.
        next

; 0= ( a -- true | false )
header code_0eq,"0=",flag_none
code_0eq:
        or      bx,bx                   ; bx = 0?
        jnz     .false                  ;
.true:  dec     bx                      ; No, return true (-1).
        next
.false: xor     bx,bx                   ; No, return false.
        next

; = ( a b -- true | false )
header code_eq,"=",flag_none
code_eq:
        pop     ax
        cmp     ax,bx                   ; NOS = TOS?
        jnz     .false                  ;
.true:  mov     bx,-1                   ; No, return true.
        next
.false: xor     bx,bx                   ; No, return false.
        next

; <> ( a b -- true | false )
header code_neq,"<>",flag_none
code_neq:
        pop     ax
        cmp     ax,bx                   ; NOS = TOS?
        jz      .false                  ;
.true:  mov     bx,-1                   ; No, return true.
        next
.false: xor     bx,bx                   ; No, return false.
        next

; < ( a b -- true | false )
header code_be,"<",flag_none
code_be:
        pop     ax
        cmp     ax,bx                   ; NOS = TOS?
        jnb     .false                  ;
.true:  mov     bx,-1                   ; No, return true.
        next
.false: xor     bx,bx                   ; No, return false.
        next

; > ( a b -- true | false )
header code_ab,">",flag_none
code_ab:
        pop     ax
        cmp     ax,bx                   ; NOS = TOS?
        jna     .false                  ;
.true:  mov     bx,-1                   ; No, return true.
        next
.false: xor     bx,bx                   ; No, return false.
        next

; @ ( a -- [a] )
header code_fetch,"@",flag_none
code_fetch:
        mov     bx,[bx]                 ; TOS = word pointed to by old TOS.
        next

; c@ ( a -- [a] )
header code_cfetch,"c@",flag_none
code_cfetch:
        mov     bl,[bx]                 ; TOS = byte pointed to by old TOS.
        mov     bh,0                    ; Fill the high byte with zeroes.
        next

; ! ( a b -- )
header code_store,"!",flag_none
code_store:
        pop     ax                      ; ax = NOS.
        mov     [bx],ax                 ; Store ax to word pointed to by TOS.
        pop     bx                      ; Pop new TOS.
        next

; +! ( a b -- )
header code_plusstore,"+!",flag_none
code_plusstore:
        pop     ax                      ; ax = NOS.
        add     [bx],ax                 ; Add ax to word pointed to by TOS.
        pop     bx                      ; Pop new TOS.
        next

; c! ( a b -- )
header code_cstore,"c!",flag_none
code_cstore:
        pop     ax                      ; ax = NOS.
        mov     [bx],al                 ; Store ax to byte pointed to by TOS.
        pop     bx                      ; Pop new TOS.
        next

; (branch) ( -- )
header code_do_branch,"(branch)",flag_none
code_do_branch:
        lodsw                           ; ax = address pointed to by IP.
        xchg    ax,si                   ; Jump to the address.
        next

; (?branch) ( a -- )
header code_do_cbranch,"(?branch)",flag_none
code_do_cbranch:
        lodsw                           ; ax = address pointed to by IP.
        or      bx,bx                   ; Is TOS == false?
        jz      .false                  ; Yes, do not jump.
        pop     bx                      ; Pop new TOS.
        mov     si,ax                   ; Jump to the address.
        next
.false:
        pop     bx                      ; Pop new TOS.
        next

; (?nbranch) ( a -- )
header code_do_cnbranch,"(?nbranch)",flag_none
code_do_cnbranch:
        lodsw                           ; ax = address pointed to by IP.
        or      bx,bx                   ; Is TOS == true?
        jnz      .false                 ; Yes, do not jump.
        pop     bx                      ; Pop new TOS.
        mov     si,ax                   ; Jump to the address.
        next
.false:
        pop     bx                      ; Pop new TOS.
        next

; dp ( -- a )
header code_dp,"dp",flag_none
header code_dp,"here",flag_none
code_dp:
        push    bx                      ; Push old TOS.
        mov     bx,[dp]                 ; TOS = dp.
        next

; dp! ( a -- )
header code_dpstore,"dp!",flag_none
code_dpstore:
        mov     [dp],bx                 ; Store TOS in dp.
        pop     bx                      ; Pop new TOS.
        next

; last ( -- a )
header code_last,"last",flag_none
code_last:
        push    bx
        mov     bx,[nptr]
        mov     bx,[bx]
        mov     bx,[bx]
        next

; last! ( a -- )
header code_laststore,"last!",flag_none
code_laststore:
        mov     di,[nptr]
        mov     di,[di]
        mov     [di],bx
        pop     bx
        next

; , ( a -- )
header code_comma,",",flag_none
code_comma:
        mov     di,[dp]                 ; di = dp.
        add     word[dp],2              ; dp += 2.
        mov     [di],bx                 ; [di] = TOS.
        pop     bx                      ; Pop new TOS.
        next

; c, ( a -- )
header code_ccomma,"c,",flag_none
code_ccomma:
        mov     di,[dp]                 ; di = dp.
        inc     word[dp]                ; dp ++.
        mov     [di],bl                 ; [di] = low byte of TOS.
        pop     bx                      ; Pop new TOS.
        next

; (
header code_comment,"(",flag_immediate
code_comment:
        call    nest
.wait:  dw      code_input              ; input
        dw      code_do_lit,')'         ; ')'
        dw      code_eq                 ; =
        dw      code_do_cnbranch,.wait  ; ?nbranch .wait
        dw      code_exit               ; exit

; (next)
header code_do_next,"(next)",flag_none
code_do_next:
        lodsw                           ; ax = address to return to if not 0.
        dec     word[bp]                ; Decrease count word.
        js      .break                  ; Below zero? Break loop.
        xchg    ax,si                   ; Above or equal to zero, loop.
        next
.break: add     bp,2                    ; Pop word from return stack.
        next

; param ( -- a )
header code_param,"param",flag_none
code_param:
        push    bx
        mov     bx,[bp]                 ; bx = pointer to mother thread.
        add     word[bp],2              ; Increase pointer.
        mov     bx,[bx]                 ; TOS = "parameter".
        next

; (does)
header code_do_does,"(does)",flag_none
code_do_does:
        xchg    bp,sp                   ; Use return stack.
        push    si                      ; Push IP.
        xchg    bp,sp                   ; Use data stack.
        pop     si                      ; Get new IP.
        pop     ax
        xchg    ax,bx
        push    ax                      ; Swap TOS and NOS.
        next

; sp0 ( -- n )
header code_sp0,"sp0",flag_none
code_sp0:
        push    bx
        mov     bx,dstack+dstack_size
        next

; sp ( -- n )
header code_sp,"sp",flag_none
code_sp:
        push    bx
        mov     bx,sp
        add     bx,2
        next

; sp! ( n -- )
header code_spstore,"sp!",flag_none
code_spstore:
        mov     sp,bx
        mov     bx,[bx-2]
        next

; rp0 ( -- n )
header code_rp0,"rp0",flag_none
code_rp0:
        push    bx
        mov     bx,rstack+rstack_size
        next

; rp ( -- n )
header code_rp,"rp",flag_none
code_rp:
        push    bx
        mov     bx,bp
        next

; rp! ( n -- )
header code_rpstore,"rp!",flag_none
code_rpstore:
        mov     bp,bx
        pop     bx
        next

; cs ( -- cs )
header code_cs,"cs",flag_none
code_cs:
        push    bx
        push    cs
        pop     bx
        next

