; graph.asm
;
; Words to work with graphics, mode 0x13 only.

; gfx-update ( -- )
; gfx-mode ( mode -- )
; gfx-sync ( -- )
; gfx-pixel13 ( x y color -- )
; gfx-xline13 ( x y length color -- )
; gfx-yline13 ( x y length color -- )
; gfx-sprite13 ( x y xsize ysize seg off -- )
; gfx-xsprite13 ( x y xsize ysize seg off -- )
; gfx-setpal ( b g r n -- )
; gfx-line13 ( x1 y1 x2 y2 color -- )
; a000 var gfx-seg

; gfx-update ( -- )
header code_gfx_update,"gfx-update",flag_none
code_gfx_update:
        push    si
        push    es
        push    ds
        mov     ax,0xa000
        mov     es,ax
        mov     ds,[var_screen]
        mov     cx,(320*200)/2
        xor     si,si
        xor     di,di
        rep     movsw
        pop     ds
        pop     es
        pop     si
        next

; gfx-mode ( -- )
header code_gfx_mode,"gfx-mode",flag_none
code_gfx_mode:
        push    bp
        push    si
        mov     ah,0x00
        mov     al,bl
        int     0x10
        pop     si
        pop     bp
        pop     bx
        next

; gfx-sync ( -- )
header code_gfx_sync,"gfx-sync",flag_none
code_gfx_sync:
        mov     dx,0x3da
.loop0: in      al,dx
        test    al,8
        jnz     .loop0
.loop1: in      al,dx
        test    al,8
        jz      .loop1
        next

; gfx-pixel13 ( x y color -- )
header code_gfx_pixel13,"gfx-pixel13",flag_none
code_gfx_pixel13:
        mov     ax,320
        pop     dx                      ; dx = y.
        mul     dx                      ; ax = 320*y
        pop     dx
        add     ax,dx                   ; ax = (320*y)+x
        xchg    ax,bx
        push    ds
        mov     dx,[var_screen]
        mov     ds,dx                   ; Set ds to video buffer segment.
        mov     [bx],al                 ; Store pixel.
        pop     ds
        pop     bx                      ; Pop new TOS.
        next

; gfx-xline13 ( x y length color -- )
header code_gfx_xline13,"gfx-xline13",flag_none
code_gfx_xline13:
        mov     ax,320
        pop     cx                      ; cx = length.
        pop     dx                      ; dx = y.
        mul     dx                      ; ax = 320*y
        pop     dx
        add     ax,dx                   ; ax = (320*y)+x
        xchg    ax,bx                   ; ax = color.
        mov     di,bx                   ; di = (320*y)+x
        push    es
        mov     dx,[var_screen]
        mov     es,dx                   ; Set es to video buffer segment.
        rep     stosb                   ; Copy data.
        pop     es
        pop     bx                      ; Pop new TOS.
        next

; gfx-yline13 ( x y length color -- )
header code_gfx_yline13,"gfx-yline13",flag_none
code_gfx_yline13:
        mov     ax,320
        pop     cx                      ; cx = length.
        pop     dx                      ; dx = y.
        mul     dx                      ; ax = 320*y
        pop     dx
        add     ax,dx                   ; ax = (320*y)+x
        xchg    ax,bx                   ; ax = color.
        mov     di,bx                   ; di = (320*y)+x
        push    es
        mov     dx,[var_screen]
        mov     es,dx                   ; Set es to video buffer segment.
.next:  stosb                           ; Copy data.
        add     di,320-1
        loop    .next
        pop     es
        pop     bx                      ; Pop new TOS.
        next

; gfx-sprite13 ( x y xsize ysize seg off -- )
header code_gfx_sprite13,"gfx-sprite13",flag_none
code_gfx_sprite13:
        pop     di
        mov     [.ds_save],ds
        pop     ds
        pop     cx
        pop     dx
        mov     ax,320
        mul     dx
        pop     dx
        add     dx,ax
        xchg    dx,di
        push    es
        mov     ax,[var_screen]
        mov     es,ax
.yloop: push    cx
.xloop: mov     al,[bx]
        inc     bx
        cmp     al,0xff
        jnz     .non_trans
        mov     al,[di]
.non_trans:
        stosb
        loop    .xloop
        pop     cx
        add     di,320
        sub     di,cx
        dec     dx
        jnz     .yloop
        pop     es
        pop     bx
        mov     ds,[cs:.ds_save]
        next
.ds_save:
        dw      0

; gfx-xsprite13 ( x y xsize ysize seg off -- )
header code_gfx_xsprite13,"gfx-xsprite13",flag_none
code_gfx_xsprite13:
        pop     di
        mov     [.ds_save],ds
        pop     ds
        pop     cx
        pop     dx
        mov     ax,320
        mul     dx
        pop     dx
        add     dx,ax
        xchg    dx,di
        push    es
        mov     ax,[var_screen]
        mov     es,ax
.yloop: push    cx
.xloop: mov     al,[bx]
        inc     bx
        mov     ah,[di]
        xor     al,ah
        stosb
        loop    .xloop
        pop     cx
        add     di,320
        sub     di,cx
        dec     dx
        jnz     .yloop
        pop     es
        pop     bx
        mov     ds,[cs:.ds_save]
        next
.ds_save:
        dw      0


; gfx-setpal ( b g r n -- )
header code_gfx_setpal,"gfx-setpal",flag_none
code_gfx_setpal:
        mov     dx,0x3c8
        xchg    ax,bx
        out     dx,al
        inc     dx
        pop     ax
        out     dx,al
        pop     ax
        out     dx,al
        pop     ax
        out     dx,al
        pop     bx
        next

; Copied from another (non-Forth) program, looks a bit ugly, but works and
; should be reasonably fast.
;
; gfx-line13 ( x1 y1 x2 y2 color -- )
;
; x1 = bp+10
; y1 = bp+8
; x2 = bp+6
; y2 = bp+4
; color = bp+2
;
; (internal use of registers:)
; cx = current x value, bx = current y value.
; si = first fb pointer, di = second fb pointer.

%define fbxincr         bp-2
%define fbyincr         bp-4
%define fbxyincr        bp-6
%define dpr             bp-8
%define dpru            bp-10
%define p               bp-12

header code_gfx_line13,"gfx-line13",flag_none
code_gfx_line13:
        push    bx                      ; "uncache" TOS
        push    bp
        mov     bp,sp
        sub     sp,12

        push    si
        push    es
        mov     ax,[var_screen]
        mov     es,ax

        mov     ax,320
        mov     dx,[bp+8]               ; y1 * SCREEN_X -> ax
        mul     dx                      ;
        add     ax,[bp+10]              ;
        xchg    ax,si                   ; ax + x1 -> si

        mov     ax,320
        mov     dx,[bp+4]               ; y2 * SCREEN_X -> ax
        mul     dx                      ;
        add     ax,[bp+6]               ;
        xchg    ax,di                   ; ax + x2 -> di

        mov     word[fbxincr],1         ; 1 -> fbxincr
        mov     cx,[bp+6]               ;
        sub     cx,[bp+10]              ; x2 - x1 -> cx
        jns     .xinit_done             ; If above zero, skip the following:
        neg     cx                      ; -cx -> cx
        mov     word[fbxincr],-1        ; -1 -> fbxincr
.xinit_done:

        mov     ax,320
        mov     word[fbyincr],ax        ; SCREEN_X -> fbyincr
        mov     bx,[bp+4]               ;
        sub     bx,[bp+8]               ; y2 - y1 -> bx
        jnc     .yinit_done             ; If above zero, skip the following:
        neg     bx                      ; -bx -> bx
        neg     ax                      ; -SCREEN_X -> ax
        mov     word[fbyincr],ax        ; ax -> fbyincr
.yinit_done:

        mov     ax,[fbxincr]            ;
        add     ax,[fbyincr]            ;
        mov     [fbxyincr],ax           ; fbxinct + fbycincr -> fbxyincr

        cmp     bx,cx                   ; y > x?
        ja      .y_independent          ;
.x_independent:
        mov     ax,bx                   ;
        add     ax,ax                   ;
        mov     [dpr],ax                ; y * 2 -> dpr 
        mov     ax,cx                   ;
        neg     ax                      ;
        mov     [p],ax                  ; -x -> p
        add     ax,ax                   ;
        mov     [dpru],ax               ; p * 2 -> dpru
        mov     bx,cx                   ;
        shr     bx,1                    ; x / 2 -> y
.xloop:

        mov     al,[bp+2]               ;
        mov     [es:si],al              ; color -> [fb1]
        mov     [es:di],al              ; color -> [fb2]

        mov     ax,[dpr]                ;
        add     [p],ax                  ; p + dpr -> p
        jns     .right_and_up1          ; if(p > 0) goto .right_and_up1

        add     si,[fbxincr]            ; fb1 + fbxincr -> fb1
        sub     di,[fbxincr]            ; fb2 - fbxincr -> fb2
        dec     bx                      ; y - 1 -> y
        jns     .xloop                  ; if(y > 0) goto .xloop

        mov     al,[bp+2]               ;
        mov     [es:si],al              ; color -> [fb1]
        test    cl,1                    ; if((x & 1) == 0) return
        jnz     .no_fix1
        jmp     .done
.no_fix1:
        mov     [es:di],al              ; color -> [fb2]
        jmp     .done

.right_and_up1:
        add     si,[fbxyincr]           ; fb1 + fbxyincr -> fb1
        sub     di,[fbxyincr]           ; fb2 - fbxyincr -> fb2
        
        mov     ax,[dpru]               ;
        add     [p],ax                  ; p + dpru -> p
        dec     bx                      ; y - 1 -> y
        jns     .xloop                  ; if(y > 0) goto .xloop

        mov     al,[bp+2]               ;
        mov     [es:si],al              ; color -> [fb1]
        test    cl,1                    ; if((x & 1) == 0) return
        jnz     .no_fix2
        jmp     .done
.no_fix2:
        mov     [es:di],al              ; color -> [fb2]
        jmp     .done

.y_independent:
        mov     ax,cx                   ;
        add     ax,ax                   ;
        mov     [dpr],ax                ; x * 2 -> dpr 
        mov     ax,bx                   ;
        neg     ax                      ;
        mov     [p],ax                  ; -y -> p
        add     ax,ax                   ;
        mov     [dpru],ax               ; p * 2 -> dpru
        mov     cx,bx                   ;
        shr     cx,1                    ; y / 2 -> x
.yloop:
        mov     al,[bp+2]               ;
        mov     [es:si],al              ; color -> [fb1]
        mov     [es:di],al              ; color -> [fb2]

        mov     ax,[dpr]                ;
        add     [p],ax                  ; p + dpr -> p
        jns     .right_and_up2          ; if(p > 0) goto .right_and_up2

        add     si,[fbyincr]            ; fb1 + fbyincr -> fb1
        sub     di,[fbyincr]            ; fb2 - fbyincr -> fb2
        dec     cx                      ; x - 1 -> x
        jns     .yloop                  ; if(x > 0) goto .xloop

        mov     al,[bp+2]               ;
        mov     [es:si],al              ; color -> [fb1]
        test    bl,1                    ; if((y & 1) == 0) return
        jz      .done
.no_fix3:
        mov     [es:di],al              ; color -> [fb2]
        jmp     .done

.right_and_up2:
        add     si,[fbxyincr]           ; fb1 + fbxyincr -> fb1
        sub     di,[fbxyincr]           ; fb2 - fbxyincr -> fb2
        mov     ax,[dpru]               ;
        add     [p],ax                  ; p + dpru -> p
        dec     cx                      ; x - 1 -> x
        jns     .yloop                  ; if(x > 0) goto .xloop

        mov     al,[bp+2]               ;
        mov     [es:si],al              ; color -> [fb1]
        test    bl,1                    ; if((y & 1) == 0) return
        jz      .done
.no_fix4:
        mov     [es:di],al              ; color -> [fb2]
.done:
        pop     es
        pop     si

        mov     sp,bp
        pop     bp

        add     sp,2*5
        pop     bx
        next

; a000 var gfx-seg
header code_gfx_seg,"gfx-seg",flag_none
code_gfx_seg:
        call    code_do_constant
var_screen:
        dw      0xa000
