; io.asm
;
; I/O-related words in the kernel are here.

; int ( intno -- )
; regs ( -- regs )
; pc@ ( port -- n )
; pc! ( n port -- )
; at ( x y -- )
; ms ( a -- )
; key ( -- a )
; key? ( -- a )
; fkey ( a -- a true | false ) <--- no a if false
; femit ( a file -- )
; fread ( file buffer length -- )
; fwrite ( file buffer length -- )
; ffread ( file seg off length -- )
; ffwrite ( file seg off length -- )
; fopen ( name mode -- file | false )
; fcreate ( name attrib -- file | false )
; fclose ( file -- )
; fseek ( file offset whence -- )
; bread ( -- )
; bkey ( -- a )
; input ( -- a )
; emit-bios ( a -- )
; emit ( a -- )
; cr ( -- )
; space ( -- )
; . ( a -- )
; type ( a count -- )
; space? ( a -- true | false )
; read ( -- )
; word ( -- a )
; .(
; print ( a -- )

; int ( intno -- )
header code_int,"int",flag_none
code_int:
        push    si
        mov     [.intno+1],bl
        mov     ax,[regs+0x00]
        mov     cx,[regs+0x02]
        mov     dx,[regs+0x04]
        mov     bx,[regs+0x06]
        mov     si,[regs+0x08]
        mov     di,[regs+0x0a]
.intno: int     0x00
        mov     [regs+0x00],ax
        mov     [regs+0x02],cx
        mov     [regs+0x04],dx
        mov     [regs+0x06],bx
        mov     [regs+0x08],si
        mov     [regs+0x0a],di
        pop     si
        pop     bx
        next

; regs ( -- regs )
header code_regs,"regs",flag_none
code_regs:
        push    bx
        mov     bx,regs
        next

regs:   times   6 dw 0

; pc@ ( port -- n )
header code_pcfetch,"pc@",flag_none
code_pcfetch:
        mov     dx,bx
        in      al,dx
        mov     bl,al
        mov     bh,0
        next

; pc! ( n port -- )
header code_pcstore,"pc!",flag_none
code_pcstore:
        pop     ax
        mov     dx,bx
        out     dx,al
        pop     bx
        next

; at ( x y -- )
header code_at,"at",flag_none
code_at:
        mov     ah,2
        pop     dx
        mov     dh,bl
        xor     bx,bx
        int     0x10
        pop     bx
        next

; ms ( a -- )
header code_ms,"ms",flag_none
code_ms:
        mov     ax,1000
        mul     bx
        xchg    ax,dx
        xchg    ax,cx                   ; cx:dx = bx*1000
        mov     ah,0x86                 ; Wait cx:dx microseconds.
        int     0x15
        pop     bx
        next

; key ( -- a )
header code_key,"key",flag_none
code_key:
        call    code_do_ref
        dw      code_key_bios

; key-bios ( -- a )
header code_key_bios,"key-bios",flag_none
code_key_bios:
        mov     ah,0x00
        int     0x16                    ; BIOS call to get ASCII char to al.
        push    bx                      ; Push old TOS.
        mov     ah,0                    ; High byte of char zeroed.
        xchg    ax,bx                   ; TOS = chatacter.
        next

; key? ( -- a )
header code_keyc,"key?",flag_none
code_keyc:
        push    bx                      ; Push old TOS.
        mov     ah,0x01
        int     0x16                    ; BIOS call to check for keypress.
        jz      .false                  ; Was there anything?
.true:  mov     bx,-1                   ; TOS = true.
        next
.false: xor     bx,bx                   ; TOS = false.
        next

; fkey ( a -- a true | false )
header code_fkey,"fkey",flag_none
code_fkey:
        mov     ah,0x3f                 ; Read from file.
        mov     cx,1                    ; One byte... Not very fast :)
        mov     dx,tmpbuf               ; Using the temporary buffer.
        int     0x21
        or      ax,ax                   ; Was anything read?
        jz      .eof                    ; Nope, EOF.
        mov     bh,0                    ; High byte of return value zero.
        mov     bl,[tmpbuf]             ; Load the byte we read to TOS.
        push    bx
        mov     bx,-1
        next
.eof:   mov     ah,0x3e                 ; EOF, close the file.
        int     0x21
        xor     bx,bx                   ; And return false.
        next

; femit ( a file -- )
header code_femit,"femit",flag_none
code_femit:
        pop     ax
        mov     [tmpbuf],al
        mov     ah,0x40
        mov     cx,1
        mov     dx,tmpbuf
        int     0x21
        pop     bx
        next

; fread ( file buffer length -- )
header code_fread,"fread",flag_none
code_fread:
        mov     cx,bx                   ; cx = TOS = length of block.
        pop     dx                      ; dx = NOS = pointer to block.
        pop     bx                      ; bx = 3rd pos = file handle.
        mov     ah,0x3f                 ; Read from file.
        int     0x21
        pop     bx                      ; Pop new TOS.
        next

; ffread ( file seg off length -- )
header code_ffread,"ffread",flag_none
code_ffread:
        mov     cx,bx                   ; cx = TOS = length of block.
        pop     dx                      ; dx = NOS = pointer to block.
        pop     ax
        pop     bx                      ; bx = 3rd pos = file handle.
        push    ds
        mov     ds,ax
        mov     ah,0x3f                 ; Read from file.
        int     0x21
        pop     ds
        pop     bx                      ; Pop new TOS.
        next

; fwrite ( file buffer length -- )
header code_fwrite,"fwrite",flag_none
code_fwrite:
        mov     cx,bx                   ; cx = TOS = length of block.
        pop     dx                      ; dx = NOS = pointer to block.
        pop     bx                      ; bx = 3rd pos = file handle.
        mov     ah,0x40                 ; Write to file.
        int     0x21
        pop     bx                      ; Pop new TOS.
        next

; ffwrite ( file seg off length -- )
header code_ffwrite,"ffwrite",flag_none
code_ffwrite:
        mov     cx,bx                   ; cx = TOS = length of block.
        pop     dx                      ; dx = NOS = pointer to block.
        pop     ax
        pop     bx                      ; bx = 3rd pos = file handle.
        push    ds
        mov     ds,ax
        mov     ah,0x40                 ; Write to file.
        int     0x21
        pop     ds
        pop     bx                      ; Pop new TOS.
        next

; fopen ( name mode -- file | false )
header code_fopen,"fopen",flag_none
code_fopen:
        mov     ah,0x3d
        mov     al,bl
        pop     dx
        int     0x21
        jc      .error
        xchg    ax,bx
        next
.error: xor     bx,bx
        next

; fcreate ( name attrib -- file | false )
header code_fcreate,"fcreate",flag_none
code_fcreate:
        mov     ah,0x3c
        mov     cx,bx
        pop     dx
        int     0x21
        jc      .error
        xchg    ax,bx
        next
.error: xor     bx,bx
        next

; fclose ( file -- )
header code_fclose,"fclose",flag_none
code_fclose:
        mov     ah,0x3e
        int     0x21
        pop     bx
        next

; fseek ( file offset whence -- )
header code_fseek,"fseek",flag_none
code_fseek:
        mov     ah,0x42
        mov     al,bl
        pop     dx
        pop     bx
        xor     cx,cx
        int     0x21
        pop     bx
        next

header code_prompt,"prompt",flag_none
code_prompt:
        call    code_do_ref
        dw      code_default_prompt
code_default_prompt:
        next

; bread ( -- )
header code_bread,"bread",flag_none
code_bread:
        call    nest
        dw      code_cr                 ; cr
        dw      code_prompt             ; prompt
        dw      code_do_lit,linebuf+1   ; linebuf
.next:  dw      code_key                ; key
        dw      code_dup                ; dup
        dw      code_do_lit,0x0d        ; 0x0d
        dw      code_eq                 ; =
        dw      code_do_cbranch,.done   ; if true, goto .done
        dw      code_dup                ; dup
        dw      code_do_lit,0x08        ; 0x08
        dw      code_eq                 ; =
        dw      code_do_cbranch,.backspace ; if true, goto .backspace
        dw      code_dup,code_emit      ; dup emit
        dw      code_over,code_cstore   ; over c!
        dw      code_1plus              ; 1+
        dw      code_do_branch,.next    ; goto .next
.done:  dw      code_drop               ; drop
        dw      code_do_lit,' '         ; ' '
        dw      code_over,code_cstore   ; over c!
        dw      code_1plus              ; 1+
        dw      code_do_lit,linebuf+1   ; linebuf+1
        dw      code_minus              ; -
        dw      code_do_lit,linebuf     ; linebuf
        dw      code_cstore             ; c!
        dw      code_space              ; space
        dw      code_exit               ; exit
.backspace:
        dw      code_over               ; over
        dw      code_do_lit,linebuf+1   ; linebuf+1
        dw      code_eq                 ; =
        dw      code_do_cnbranch,.bs_ok ; if false, goto .bs_ok
        dw      code_drop
        dw      code_do_branch,.next    ; goto .next
.bs_ok:
        dw      code_emit               ; emit
        dw      code_do_lit,' '         ; ' '
        dw      code_emit               ; emit
        dw      code_do_lit,0x08        ; 0x08
        dw      code_emit               ; emit
        dw      code_1minus             ; 1-
        dw      code_do_branch,.next    ; goto .next

; bkey ( -- a )
header code_bkey,"bkey",flag_none
code_bkey:
        call    nest
.next:  dw      code_do_lit,linebuf     ; linebuf
        dw      code_cfetch             ; c@
        dw      code_do_lit,bptr        ; bptr
        dw      code_fetch              ; @
        dw      code_eq                 ; =
        dw      code_do_cbranch,.eol    ; if true, goto .eol
        dw      code_do_lit,bptr        ; bptr
        dw      code_fetch              ; @
        dw      code_1plus,code_dup     ; 1+ dup
        dw      code_do_lit,bptr        ; bptr
        dw      code_store              ; !
        dw      code_do_lit,linebuf     ; linebuf
        dw      code_plus,code_cfetch   ; + c@
        dw      code_exit               ; exit
.eol:   dw      code_bread              ; bread
        dw      code_do_lit,0           ; 0
        dw      code_do_lit,bptr        ; bptr
        dw      code_store              ; !
        dw      code_do_lit,0xa         ; 0xa
        dw      code_exit               ; exit

; input ( -- a )
header code_input,"input",flag_none
code_input:
        call    code_do_ref
        dw      code_input_std

; input-std ( -- a )
header code_input_std,"input-std",flag_none
code_input_std:
        call    nest
.begin: dw      code_do_lit,fptr        ; fptr
        dw      code_fetch,code_fetch   ; @ @
        dw      code_dup                ; dup
        dw      code_do_cbranch,.file   ; if not 0, goto .file
        dw      code_drop               ; drop
        dw      code_bkey               ; bkey
        dw      code_exit               ; exit
.file:  dw      code_fkey               ; fkey    (read byte from file)
        dw      code_do_cnbranch,.eof   ; if false, goto .eof
        dw      code_exit               ; exit
.eof:   dw      code_do_lit,2           ; 2
        dw      code_do_lit,fptr        ; fptr
        dw      code_plusstore          ; +!
        dw      code_do_lit,' '         ; ' '   (return space on EOF)
        dw      code_exit               ; exit

; emit-bios ( a -- )
header code_emit_bios,"emit-bios",flag_none
code_emit_bios:
        mov     ah,0x0e
        mov     al,bl
        int     0x10                    ; Print TOS to screen.
        pop     bx                      ; Pop TOS.
        next

; emit ( a -- )
header code_emit,"emit",flag_none
code_emit:
        call    code_do_ref
        dw      code_emit_bios

emit_al:
        push    bx                      ; Now bx returns to its old value.
        mov     bl,al                   ; Set TOS to character.
        hl_call code_emit               ; Call emit.
        ret

; cr ( -- )
header code_cr,"cr",flag_none
code_cr:
        mov     al,0x0d
        call    emit_al                 ; cr
        mov     al,0x0a
        call    emit_al                 ; lf
        next

; space ( -- )
header code_space,"space",flag_none
code_space:
        mov     al,' '
        call    emit_al                 ; Print space.
        next

; . ( a -- )
header code_dot,".",flag_none
code_dot:
        mov     cx,[base]               ; cx = numeric base.
        mov     di,tmpbuf               ; di = pointer to temporary buffer.
.next:  mov     ax,bx                   ; ax = number.
        xor     dx,dx                   ; dx:ax = ax.
        div     cx                      ; ax = dx:ax/cx, dx = dx:ax mod cx.
        mov     bx,ax                   ; Quotient becomes new number.
        sub     dl,0x0a                 ; Subtract 10 from dl.
        jc      .digit                  ; Is carry flag set? It's a digit.
        add     dl,'a'                  ; Otherwise, a character.
        jmp     short .char
.digit: add     dl,0x0a+'0'             ; A decimal digit...
.char:  mov     [di],dl                 ; Digit/character stored in string.
        inc     di                      ;
        or      bx,bx                   ; Is the number 0?
        jnz     .next                   ; No, repeat.

.print: dec     di                      ; di--.
        mov     al,[di]                 ; al = [di].
        push    di                      ; Save di.
        call    emit_al                 ; Print al.
        pop     di                      ; Restore di
        cmp     di,tmpbuf               ; Back at the start of the string?
        jnz     .print                  ; No, print next character.
        pop     bx
        next

; type ( a count -- )
header code_type,"type",flag_none
code_type:
        mov     cx,bx                   ; cx = count.
        pop     di                      ; di = string to print.
.next:  mov     al,[di]                 ; al = char to print.
        inc     di
        push    cx
        push    di
        call    emit_al                 ; Print al.
        pop     di
        pop     cx
        loop    .next                   ; Loop.
        pop     bx                      ; Pop TOS.
        next

; space? ( a -- true | false )
header code_is_space,"space?",flag_none
code_is_space:
        cmp     bl,' '                  ; Space?
        jz      .true
        cmp     bl,0x0d                 ; cr?
        jz      .true
        cmp     bl,0x0a                 ; lf?
        jz      .true
        cmp     bl,0x09                 ; Tab?
        jz      .true
        xor     bx,bx                   ; Niether, return false.
        next
.true:  mov     bx,-1                   ; Return true.
        next

; read ( -- )
header code_read,"read",flag_none
code_read:
        call    nest
.wait1: dw      code_input              ; input
        dw      code_dup,code_is_space  ; dup space?
        dw      code_0eq                ; 0=
        dw      code_do_cbranch,.done1  ; if true, goto .done1
        dw      code_drop               ; drop
        dw      code_do_branch,.wait1   ; goto .wait1
.done1: dw      code_do_lit,wordbuf+1   ; wordbuf+1
        dw      code_cstore             ; c!
        dw      code_do_lit,wordbuf+2   ; wordbuf+2
.wait2: dw      code_input              ; input
        dw      code_dup,code_is_space  ; dup space?
        dw      code_do_cbranch,.done2  ; if true, goto .done2
        dw      code_over,code_cstore   ; over c!
        dw      code_1plus              ; 1+
        dw      code_do_branch,.wait2   ; goto .wait2
.done2: dw      code_drop               ; drop
        dw      code_do_lit,wordbuf+1   ; wordbuf+1
        dw      code_minus              ; -
        dw      code_do_lit,wordbuf     ; wordbuf
        dw      code_cstore             ; c!
        dw      code_exit               ; exit

; word ( -- a )
header code_word,"word",flag_none
code_word:
        push    bx                      ; Push TOS.
        mov     bx,wordbuf              ; Address to wordbuf becomes new TOS.
        next

; .(
header code_dotparen,".(",flag_immediate
code_dotparen:
        call    nest
.next:  dw      code_input,code_dup     ; input dup
        dw      code_do_lit,')'         ; ')'
        dw      code_eq                 ; =
        dw      code_do_cbranch,.done   ; ?branch .done
        dw      code_emit               ; emit
        dw      code_do_branch,.next    ; branch next
.done:  dw      code_drop,code_exit     ; drop exit

; print ( a -- )
header code_print,"print",flag_none
code_print:
        call    nest
        dw      code_count,code_type    ; count type
        dw      code_exit               ; exit
