; mem.asm
;
; Functions for memory managment.

; far! ( n offset segment -- )
; farc! ( n offset segment -- )
; far@ ( offset segment -- n )
; farc@ ( offset segment -- n )
; malloc ( blocks -- segment | false )
; mfree ( segment -- )
; segment ( -- segment )
; copy ( src dest bytes -- )
; copyw ( src dest words -- )
; mset ( byte address count -- )
; msetw ( word address count -- )
; farcopy ( src-segment dest-segment bytes -- )
; farcopyw ( src-segment dest-segment words -- )

; far! ( n offset segment -- )
header code_farstore,"far!",flag_none
code_farstore:
        pop     di                      ; di = offset.
        pop     ax                      ; ax = n.
        push    ds                      ; Save data segment.
        mov     ds,bx                   ; ds = segment.
        mov     [di],ax                 ; Store word.
        pop     ds                      ; Restore data segment.
        pop     bx                      ; Pop new TOS.
        next

; farc! ( n offset segment -- )
header code_farcstore,"farc!",flag_none
code_farcstore:
        pop     di                      ; di = offset.
        pop     ax                      ; ax = n.
        push    ds                      ; Save data segment.
        mov     ds,bx                   ; ds = segment.
        mov     [di],al                 ; Store byte.
        pop     ds                      ; Restore data segment.
        pop     bx                      ; Pop new TOS.
        next

; far@ ( offset segment -- n )
header code_farfetch,"far@",flag_none
code_farfetch:
        pop     di                      ; di = offset.
        push    ds                      ; Save data segment.
        mov     ds,bx                   ; ds = segment.
        mov     bx,[di]                 ; Fetch word.
        pop     ds                      ; Restore data segment.
        next

; farc@ ( offset segment -- n )
header code_farcfetch,"farc@",flag_none
code_farcfetch:
        pop     di                      ; di = offset.
        push    ds                      ; Save data segment.
        mov     ds,bx                   ; ds = segment.
        mov     bl,[di]                 ; Fetch word.
        mov     bh,0                    ; Zero high byte of TOS.
        pop     ds                      ; Restore data segment.
        next

; malloc ( blocks -- segment | false )
header code_malloc,"malloc",flag_none
code_malloc:
        mov     ah,0x48                 ; Allocate memory.
        int     0x21                    ; bx already is the number of blocks.
        jc      .error                  ; Error?
        xchg    ax,bx                   ; No, return segment.
        next
.error: xor     bx,bx                   ; Yes, return false.
        next

; mfree ( segment -- )
header code_mfree,"mfree",flag_none
code_mfree:
        mov     ah,0x49                 ; Free memory.
        push    es
        mov     es,bx                   ; es = TOS = segment.
        int     0x21
        pop     es
        pop     bx                      ; Pop new TOS.
        next

; segment ( -- segment )
header code_segment,"segment",flag_none
code_segment:
        push    bx
        mov     bx,cs
        next

; copy ( src dest bytes -- )
header code_copy,"copy",flag_none
code_copy:
        mov     cx,bx                   ; cx = number of bytes to copy.
        mov     bx,si                   ; Backup IP.
        pop     di                      ; dest
        pop     si                      ; src
        rep     movsb                   ; Copy data.
        mov     si,bx                   ; Restore IP.
        pop     bx                      ; Pop new TOS.
        next

; copyw ( src dest words -- )
header code_copyw,"copyw",flag_none
code_copyw:
        mov     cx,bx                   ; cx = number of words to copy.
        mov     bx,si                   ; Backup IP.
        pop     di                      ; dest
        pop     si                      ; src
        rep     movsw                   ; Copy data.
        mov     si,bx                   ; Restore IP.
        pop     bx                      ; Pop new TOS.
        next

; mset ( byte address count -- )
header code_mset,"mset",flag_none
code_mset:
        mov     cx,bx
        pop     di
        pop     ax
        rep     stosb
        pop     bx
        next

; msetw ( word address count -- )
header code_msetw,"msetw",flag_none
code_msetw:
        mov     cx,bx
        pop     di
        pop     ax
        rep     stosw
        pop     bx
        next

; farcopy ( src-segment dest-segment bytes -- )
header code_farcopy,"farcopy",flag_none
code_farcopy:
        mov     cx,bx                   ; cx = number of bytes to copy.
        pop     ax                      ; dest-segment
        pop     dx                      ; src-segment

        push    ds
        push    es
        push    si
        xor     si,si
        xor     di,di
        mov     ds,dx                   ; ds = src.
        mov     es,ax                   ; es = dest.
        rep     movsb                   ; Copy block.
        pop     si
        pop     es
        pop     ds

        pop     bx                      ; Pop new TOS.
        next

; farcopyw ( src-segment dest-segment words -- )
header code_farcopyw,"farcopyw",flag_none
code_farcopyw:
        mov     cx,bx                   ; cx = number of words to copy.
        pop     ax                      ; dest-segment
        pop     dx                      ; src-segment

        push    ds
        push    es
        push    si
        xor     si,si
        xor     di,di
        mov     ds,dx                   ; ds = src.
        mov     es,ax                   ; es = dest.
        rep     movsw                   ; Copy block.
        pop     si
        pop     es
        pop     ds

        pop     bx                      ; Pop new TOS.
        next
