; string.asm
;
; String-related words in the kernel should be in here.

; scopy ( src dest -- )
; s= ( a b -- true | false )
; >asciiz ( a -- )
; s, ( a -- )
; s>n ( a -- n true | false ) <-- no n if false.
; count ( a -- a+1 count )

; scopy ( src dest -- )
header code_scopy,"scopy",flag_none
code_scopy:
        pop     di
        mov     cl,[di]
        mov     ch,0
        inc     cx                      ; ax = length + 1 (to include count)
.next:  mov     al,[di]
        mov     [bx],al
        inc     di
        inc     bx
        loop    .next
        pop     bx
        next

; s= ( a b -- true | false )
header code_seq,"s=",flag_none
code_seq:
        pop     di                      ; di = string a, bx = string b.
        mov     al,[bx]                 ; al = length of string b.
        cmp     al,[di]                 ; Equal to length of string a?
        jnz     .false                  ; No, return false.
.next:  inc     bx                      ; Go to next character of string b.
        inc     di                      ; Go to next character of string a.
        mov     ah,[bx]                 ; ah is a character in string b.
        cmp     ah,[di]                 ; Compare to corresponding char in a.
        jnz     .false                  ; Not equal? Return false.
        dec     al                      ; Was this the last character pair?
        jnz     .next                   ; No, compare another pair.
.true:  mov     bx,-1                   ; Equal! Return true.
        next
.false: xor     bx,bx
        next

; >asciiz ( a -- )
header code_toasciiz,">asciiz",flag_none
code_toasciiz:
        mov     cl,[bx]
        mov     ch,0                    ; cx = length of string.
        mov     dx,cx                   ; dx = cx.
.next:  mov     al,[bx+1]
        mov     [bx],al
        inc     bx
        loop    .next
        mov     byte[bx],0
        pop     bx
        next

; s, ( a -- )
header code_scomma,"s,",flag_none
code_scomma:
        mov     di,[dp]                 ; di = dp.
        mov     cl,[bx]
        mov     ch,0                    ; cx = length of string.
        mov     [di],cl                 ; Store length.
.next:  inc     bx
        inc     di
        mov     al,[bx]                 ; al = character.
        mov     [di],al                 ; Store to heap.
        loop    .next
        inc     di
        mov     [dp],di                 ; Store new dp.
        pop     bx
        next

; s>n ( a -- n true | false ) <-- no n if false.
header code_ston,"s>n",flag_none
code_ston:
        mov     byte[.neg_flag],0
        mov     cl,[bx]
        mov     ch,0                    ; cx = length of string.
        xor     di,di                   ; di = number.
        cmp     byte[bx+1],'-'
        jnz     .next
        inc     bx
        dec     cx
        mov     byte[.neg_flag],1
.next:  inc     bx
        mov     al,[bx]                 ; al = character.
        cmp     al,'0'
        jb      .not_digit
        cmp     al,'9'
        ja      .not_digit
.digit:
        sub     al,'0'
.process:
        cmp     al,[base]               ; Digit below base?
        jnb     .error                  ; No, error.
        mov     ah,0                    ; ax = digit.
        push    ax
        mov     ax,[base]               ; ax = base.
        mul     di                      ; ax = base * number.
        pop     dx
        xchg    ax,di
        add     di,dx                   ; number = (base * number) + digit.

        loop    .next
.done:  cmp     byte[.neg_flag],0
        jz      .not_neg
        neg     di
.not_neg:
        push    di                      ; Save number.
        mov     bx,-1                   ; Return true.
        next
.not_digit:
        cmp     al,'a'
        jb      .error
        cmp     al,'z'
        ja      .error
.char:
        sub     al,'a'-0x0a
        jmp     short .process
.error:
        xor     bx,bx                   ; Return false.
        next

.neg_flag       db      0

; count ( a -- a+1 count )
header code_count,"count",flag_none
code_count:
        mov     al,[bx]                 ; al = length of string.
        inc     bx                      ; a++.
        push    bx                      ; Push a.
        mov     ah,0                    ; High byte of count = 0.
        xchg    ax,bx                   ; Set TOS to count word.
        next

