; TFORTH
; A simple DOS Forth by Robert Oestling
; http://robert.zizi.org -- robost(at)home(dot)se
;
; Header.
;
; 00    word    Pointer to last header.
; 02    word    Pointer to code.
; 04    byte    Flags.
; 05    byte    Name length.
; 06            Name string.
; --    word    Pointer to start of header.
;
; Flags.
;
; 0             Immediate.
;
; Register usage.
;
; ss:sp         Data stack pointer (SP).
; ss:bp         Return stack pointer (RP).
; ds:si         Interpretation pointer (IP).
; bx            TOS.
;

%define rstack_size     0x40*2
%define dstack_size     0x40*2
%define fstack_size     0x10*2

%include "macros.asm"

        org     0x100

        cld

        mov     ah,0x4a                 ; Resize our memory block.
        mov     bx,0x1000               ; 64kB
        int     0x21

        mov     word[fstack+fstack_size-2],0
        mov     byte[linebuf],0

        mov     bp,rstack+rstack_size
        mov     sp,dstack+dstack_size
        mov     si,code_bye

        mov     word[mode],mode_interpret
        mov     word[base],0x10
        mov     word[fptr],fstack+fstack_size-2
        mov     word[bptr],0
        jmp     short code_quit

        xor     bx,bx

header  code_quit,"quit",flag_none
code_quit:
        call    code_do_ref
        dw      code_default_quit

code_default_quit:
        call    nest
        dw      code_do_lit,.mainf
        dw      code_fload_file         ; fload main.f
.next:
        dw      code_read,code_word     ; Read a word.
        dw      code_process            ; Process it.
        dw      code_drop               ; We don't want to know the result.
        dw      code_do_branch,.next    ; Repeat.
.mainf  db      "main.f",0

; You can NOT do without these:
%include "core.asm"
%include "io.asm"
%include "string.asm"
%include "fload.asm"
%include "compiler.asm"

; You can do without these:
%include "mem.asm"
%include "gfx.asm"
%include "mouse.asm"

; 0x10 var base
header  code_base,"base",flag_none
code_base:
        call    code_do_constant
base    dw      0x10

; nstack.end var nptr
header  code_nptr,"nptr",flag_none
code_nptr:
        call    code_do_constant
nptr    dw      nstack.last

nstack:
        times   0x0f dw 0
.last:  dw      core_last
.end:

core_last:
        dw      last_header

mode    dw      mode_interpret
dp      dw      heap
fptr    dw      fstack+fstack_size-2
bptr    dw      0

[section .bss]

rstack  resb    rstack_size
dstack  resb    dstack_size
fstack  resb    fstack_size

linebuf resb    0x80
tmpbuf  resb    0x20
wordbuf resb    0x20

heap:
